#-----------------------------------------------------------------------------
#
#   PyWeek 10 - Game
#
#-----------------------------------------------------------------------------

from __future__ import division
from math import ceil, pi, sin
from time import time
from numpy import zeros, ones, array, add
from OpenGL import GL
import pygame
from pygame import Surface, SRCALPHA
from pygame.surfarray import array_alpha
from fixes import pixels_alpha
from albow.resource import get_image
from humerus.hgame import HGame
import perf
from options import options
from texture import Texture
from vectors import Point, Vector, zero_vector
from sprite import Sprite, Player, FACE_LEFT, FACE_RIGHT, FACE_FRONT, \
	ACT_STAND, ACT_WALK, ACT_SONIC, \
	DIR_LEFT, DIR_RIGHT, DIR_UP, DIR_DOWN
import level
from texture import pow2up

#-----------------------------------------------------------------------------

#grid_cell_size = 32
grid_cell_size = 8
R = 32 / grid_cell_size
Kh = 2.0 * R          # Spring stretch force constant
Kv = 0.5 * R          # Spring flex force constant
M = 0.25 / R          # Mass at each grid point
D = 0.99							# Velocity damping factor
Ff = 50.0 * R         # Forcing amplitude
ff = 0.1              # Forcing frequency

sonic_dematerialisation_factor = 0.98
sonic_obliteration_threshold = 0.2
steadiness_threshold = 5.0
game_completion_steady_time = 5.0 # seconds

wf = 2 * pi * ff

class SpringGrid(object):
	#  dimensions   (w, h)                Number of cells in grid
	#  P            float[w+1, h+1, 2]    Positions of grid points
	#  V            float[w+1, h+1, 2]    Velocities of grid points
	#  t            float                 Phase of forcing function
	#  solidity     float[w, h]           0.0 to 1.0
	#  vertices     float[:, :, 2]        Vertex array for GL (view of P)
	#  tex_coords   float[:, :, 2]        Tex coords array for GL
	#  indices      float[:]              Array indices for GL

	def __init__(self, image, texture_size):
		size = image.get_size()
		pw, ph = size
		w = int(ceil(pw / grid_cell_size))
		h = int(ceil(ph / grid_cell_size))
		self.dimensions = (w, h)
		self.init_physics()
		self.init_solidity(image)
		self.init_vertex_arrays(texture_size)
		self.t = 0.0
	
	def init_physics(self):
		w, h = self.dimensions
		P = self.P = zeros((w + 1, h + 1, 2))
		V = self.V = zeros((w + 1, h + 1, 2))
		for i in xrange(w + 1):
			P[i, :, 0] = i * grid_cell_size
		for j in xrange(h + 1):
			P[:, j, 1] = j * grid_cell_size
	
	def init_solidity(self, image):
		idim, jdim = self.dimensions
		pw, ph = image.get_size()
		mask = zeros(self.dimensions)
		alpha = array_alpha(image)
		for j in xrange(jdim):
			y = int(ph * (j + 0.5) / jdim)
			for i in xrange(idim):
				x = int(pw * (i + 0.5) / idim)
				mask[i, j] = alpha[x, ph - y] / 255.0
		self.solidity = mask
		if 0:
			print "GameState: solidity:"
			for j in xrange(jdim):
				for i in xrange(idim):
					print "%3.1f" % mask[i, j],
				print

	def init_vertex_arrays(self, texture_size):
		w, h = self.dimensions
		tw, th = texture_size
		us = grid_cell_size / tw
		ut = grid_cell_size / th
		s = self.solidity
		tex_coords = zeros((w + 1, h + 1, 2))
		indices = []
		def v(i, j):
			return i * (h + 1) + j
		for j in xrange(h + 1):
			for i in xrange(w + 1):
				tex_coords[i, j] = (i * us, j * ut)
		for j in xrange(h):
			for i in xrange(w):
				if s[i, j]:
					indices.extend([v(i, j), v(i+1, j), v(i+1, j+1), v(i, j+1)])
		self.vertices = self.P.reshape((-1, 2))
		self.tex_coords = tex_coords.reshape((-1, 2))
		self.indices = array(indices, 'i')
	
	def timestep(self, dt):
		w, h = self.dimensions
		P = self.P
		V = self.V
		S = self.solidity
		L = grid_cell_size                          # Resting length of horz springs
		F = zeros((w + 1, h + 1))                   # Forces on grid points
		Sh = zeros((w, h + 1))                      # Horz solidity
		Sv = zeros((w + 1, h))                      # Vert solidity
		Sh[:, :-1] = S
		Sh[:, 1:] += S
		Sv[:-1, :] = S
		Sv[1:, :] = S
		Fh = Kh * Sh * (P[1:, :, 0] - P[:-1, :, 0] - L)  # Horz spring stretch force
		Fv = Kv * Sv * (P[:, 1:, 0] - P[:, :-1, 0])      # Vert spring flex force
		F[:-1, :] += Fh
		F[1:, :] -= Fh
		F[:, :-1] += Fv
		F[:, 1:] -= Fv
		if not game.state.frequency_locked:
			F[-1, -1] += Ff * sin(wf * self.t)
		V *= D
		V[..., 0] += F * (dt / M)
		V[:, 0] = 0.0
		P += V * dt
		self.t += dt
		#print "SpringGrid.timestep: V[-1, -1] =", V[-1, -1] ###
		if abs(V[-1, -1, 0]) <= steadiness_threshold:
			game.state.steady_time += dt
		else:
			game.state.steady_time = 0.0
		#print "...steady for", game.state.steady_time ###
	
#	def world_to_grid(self, wp):
#		return wp / grid_cell_size

	def step_present(self, i, j):
		#print "step_present:", i, j ###
		idim, jdim = self.dimensions
		s = self.solidity
		return 0 <= i < idim \
			and 1 <= j <= jdim \
			and s[i, j-1] >= 0.1 \
			and s[i, j] < 0.1

	def cells_covered_by_rect(self, rect):
		left, bottom, right, top = rect
		idim, jdim = self.dimensions
		j0 = max(0, int(round(bottom / grid_cell_size)))
		j1 = max(j0, min(jdim, int(round(top / grid_cell_size))))
		P = self.P
		s = self.solidity
		for j in xrange(j0, j1):
			for i in xrange(idim):
				if s[i, j]:
					x0 = max(left, 0.5 * (P[i, j, 0] + P[i, j+1, 0]))
					x1 = min(right, 0.5 * (P[i+1, j, 0] + P[i+1, j+1, 0]))
					if x0 < x1:
						yield i, j

	def collide_rect(self, left, bottom, right, top, facing):
		#print "collide_rect:", left, bottom, right, top ###
		#ref = [left, right][facing]
		idim, jdim = self.dimensions
		j0 = max(0, int(round(bottom / grid_cell_size)))
		j1 = max(j0, min(jdim, int(round(top / grid_cell_size))))
		#print "j0 =", j0, "j1 =", j1 ###
		P = self.P
		s = self.solidity
		overlaps = zeros(j1 - j0)
		for j in xrange(j0, j1):
			#print "j =", j ###
			for i in xrange(idim):
				if s[i, j] >= 0.1:
					x0 = max(left, 0.5 * (P[i, j, 0] + P[i, j+1, 0]))
					x1 = min(right, 0.5 * (P[i+1, j, 0] + P[i+1, j+1, 0]))
					#if x0 < right and x1 > left:
					if x0 < x1:
						#print "i =", i, "j =", j, "x0 =", x0, "x1 =", x1 ###
						#d0 = x0 - ref
						#d1 = x1 - ref
						#overlaps[j-j0] += abs(d1 * d1 - d0 * d0)
						overlaps[j-j0] += 1
		#print overlaps ###
		return overlaps

	def find_anchor(self, world_pos):
		#  Given a sprite foot position in world coordinates, if the foot
		#  is immediately above a solid cell, returns (i, j, u) where i, j
		#  are the cell indices and 0.0 <= u <= 1.0 is the horizontal distance
		#  from the left of the cell as a fraction of the cell size.
		#  Otherwise, returns None.
		wx, wy = world_pos
		idim, jdim = self.dimensions
		P = self.P
		s = self.solidity
		j = int(round(wy / grid_cell_size))
		if 1 <= j <= jdim:
			for i in xrange(idim):
				if s[i, j-1] >= 0.1:
					x0 = P[i, j, 0]
					x1 = P[i+1, j, 0]
					if x0 <= wx <= x1:
						return (i, j, (wx - x0) / (x1 - x0))
	
	def anchor_position(self, anchor):
		#  Given a tuple returned by find_anchor, returns the current position
		#  in world coordinates corresponding to the anchor point.
		i, j, u = anchor
		P = self.P
		x0 = P[i, j, 0]
		x1 = P[i+1, j, 0]
		wx = x0 + u * (x1 - x0)
		wy = j * grid_cell_size
		return Point(wx, wy)

	def clear_above_anchor(self, anchor):
		if anchor:
			i, j = anchor[:2]
			s = self.solidity[i-1:i+2, j:j+2]
			m = s < 0.1
			return m.all()
		else:
			return True
	
	def sonic_cell(self, indices):
		i, j = indices
		S = self.solidity
		s = S[indices] * sonic_dematerialisation_factor
		if s < sonic_obliteration_threshold:
			s = 0.0
		S[indices] = s
		#print "State.sonic_cell:", indices, "solidity now", s ###
		g = grid_cell_size
		x = i * g
		y = j * g
		jelly = game.jelly
		h = jelly.get_height()
		surf = jelly.subsurface((x, h - y - g, g, g))
		#print "surf masks: %08x %08x %08x %08x" % surf.get_masks()
		alpha = pixels_alpha(surf)
		alpha[:] = s * 255
		data = pygame.image.tostring(surf, 'RGBA', True)
		GL.glBindTexture(GL.GL_TEXTURE_2D, game.jelly_texture)
		GL.glTexSubImage2D(GL.GL_TEXTURE_2D, 0, x, y, g, g, GL.GL_RGBA,
			GL.GL_UNSIGNED_BYTE, data)

#-----------------------------------------------------------------------------

class GameState(object):
	#  grid       SpringGrid
	#  sprites    [Sprite]
	
	scroll_offset = 0
	target_scroll_offset = 0
	microwave_frequency = 1039
	target_microwave_frequency = 1210
	frequency_lock_range = 10
	frequency_locked = False
	steady_time = 0.0
	
	def __init__(self):
		self.grid = SpringGrid(game.jelly, game.jelly_texture_size)
		self.sprites = []
		self.init_sprites()
	
	def init_sprites(self):
		level.add_furniture(self)
		player = Player()
		self.player = player
		self.add_sprite(player, level.player_start_position)
		y = player.position[1]
		if y > 320:
			self.scroll_offset = self.target_scroll_offset = y - 64

	def add_sprite(self, sprite, world_pos):
		self.sprites.append(sprite)
		sprite.position = Point(*world_pos)
	
	def begin_frame(self):
		dt = 1 / options.frame_rate_option
		self.update_physics(dt)
	
	def update_physics(self, dt):
		f = self.microwave_frequency
		ft = self.target_microwave_frequency
		self.frequency_locked = abs(f - ft) <= self.frequency_lock_range
		for sprite in self.sprites:
			self.anchor_sprite(sprite)
		self.grid.timestep(dt)
		for sprite in self.sprites:
			sprite.update_physics(game, dt)

	def anchor_sprite(self, sprite):
		grid = self.grid
		anchor = None
		if sprite.is_stepped_up():
			anchor = grid.find_anchor(sprite.stepped_up_position())
			if anchor:
				sprite.finish_stepping_up()
		if not anchor:
			anchor = grid.find_anchor(sprite.position)
		sprite.anchor = anchor
		if anchor:
			sprite.velocity = zero_vector
	
	def move_sprite(self, sprite, vector):
		#print "move_sprite:", vector ###
		grid = self.grid
		p0 = sprite.stepped_up_position()
		x0, y0 = p0
		p = p0 + vector
		x, y = p
		cw, ch = sprite.collision_extent
		anchor = sprite.anchor
		if anchor:
			sy = sprite.available_step_up()
		else:
			sy = 0.0
		dir = vector[0] > 0.0
		init_overlaps = grid.collide_rect(x0 - cw, y0, x0 + cw, y0 + ch, dir)
		init_overlap = add.reduce(init_overlaps)
		#print "init_overlaps =", init_overlaps, "total", init_overlap ###
		overlaps = grid.collide_rect(x - cw, y, x + cw, y + ch + sy, dir)
		#print "overlaps =", overlaps ###
		jh = int(round(ch / grid_cell_size))
		#print "jh =", jh ###
		for k in xrange(len(overlaps) - jh + 1):
			k_overlap = add.reduce(overlaps[k:k+jh])
			#print "overlap at k=", k, "is", k_overlap ###
			if k_overlap <= init_overlap:
				#print "space at k =", k ###
				if k == 0:
					sprite.position = sprite.position + vector
					return True
				elif anchor:
					i0, j0 = anchor[:2]
					f = sprite.facing_sign()
					di = int(round((x + f * cw - x0) / grid_cell_size))
					if grid.step_present(i0 + di, j0 + k):
						sprite.begin_stepping_up(k * grid_cell_size)
						break

	def sonic(self, origin, rect):
		#print "GameState.sonic:", origin, rect ###
		for sprite in self.sprites:
			sprite.sonic(game, origin)
		grid = self.grid
		for indices in grid.cells_covered_by_rect(rect):
			grid.sonic_cell(indices)
	
	def get_locked_frequency(self):
		locked = self.frequency_locked
		if locked:
			f = self.target_microwave_frequency
		else:
			f = self.microwave_frequency
		return f, locked

#-----------------------------------------------------------------------------

class Game(HGame):
	#  action      ACT_*     Continuous action most recently commanded
	#  jump_flag   boolean   Jump commanded

	def __init__(self, *args, **kwds):
		HGame.__init__(self, *args, **kwds)
		jelly = self.jelly = get_image("jelly_layer.png")
		w, h = jelly.get_size()
		w2 = pow2up(w)
		h2 = pow2up(h)
		self.jelly_texture_size = (w2, h2)
		data = pygame.image.tostring(jelly, 'RGBA', True)
		#data = jelly
		t = self.jelly_texture = GL.glGenTextures(1)
		GL.glBindTexture(GL.GL_TEXTURE_2D, t)
		GL.glTexImage2D(GL.GL_TEXTURE_2D, 0,
			GL.GL_RGBA, w2, h2, 0,
			GL.GL_RGBA, GL.GL_UNSIGNED_BYTE, None)
		GL.glTexSubImage2D(GL.GL_TEXTURE_2D, 0,
			0, 0, w, h,
			GL.GL_RGBA, GL.GL_UNSIGNED_BYTE, data)
		self.sounds_playing = set()
	
	def start_sound(self, sound, **kwds):
		sounds = self.sounds_playing
		if sound not in sounds:
			sounds.add(sound)
			sound.play(**kwds)
	
	def stop_sound(self, sound):
		sound.stop()
		self.sounds_playing.discard(sound)
	
	def get_player(self):
		return self.state.player
	
	player = property(get_player)
	
	def new_game(self):
		HGame.new_game(self)
		self.action = ACT_STAND
		self.jump_flag = False
	
	def new_state(self):
		return GameState()
	
	def game_is_completed(self):
		return self.state.steady_time >= game_completion_steady_time
	
	def begin_frame(self):
		#t1 = time()
		perf.begin("Physics")
		self.state.begin_frame()
		#t2 = time()
		#perf.used_time("Physics", t2 - t1)
		perf.end("Physics")
	
	def set_action_and_facing(self, action, facing):
		player = self.player
		if self.action <> action or (facing is not None and player.facing <> facing):
			player.cancel_step_up()
			self.action = action
			if facing is not None:
				player.facing = facing
			if action == ACT_WALK:
				player.reset_walk_phase()
	
	def start_sonic(self, dir, facing):
		self.set_action_and_facing(ACT_SONIC, facing)
		self.player.sonic_dir = dir
	
	def start_walking_left(self):
		self.set_action_and_facing(ACT_WALK, FACE_LEFT)

	def start_walking_right(self):
		self.set_action_and_facing(ACT_WALK, FACE_RIGHT)
	
	def stand_still(self):
		self.set_action_and_facing(ACT_STAND, None)
	
	def face_front(self):
		self.set_action_and_facing(ACT_STAND, FACE_FRONT)
	
	def stop_walking_left(self):
		if self.action == ACT_WALK and self.player.facing == FACE_LEFT:
			self.stand_still()
	
	def stop_walking_right(self):
		if self.action == ACT_WALK and self.player.facing == FACE_RIGHT:
			self.stand_still()
	
	def jump(self):
		self.jump_flag = True
	
	def cancel_jump(self):
		self.jump_flag = False
	
	def sonic_left(self):
		self.start_sonic(DIR_LEFT, FACE_LEFT)

	def sonic_right(self):
		self.start_sonic(DIR_RIGHT, FACE_RIGHT)

	def sonic_up(self):
		self.start_sonic(DIR_UP, FACE_FRONT)

	def sonic_down(self):
		self.start_sonic(DIR_DOWN, FACE_FRONT)
		
	def stop_sonic(self):
		if self.action == ACT_SONIC:
			self.action = ACT_STAND

game = Game()
